module HTH
  class UserSyncher

    attr_accessor :ldap
    attr_accessor :rest
    attr_accessor :errors
    attr_accessor :api_url

    def initialize(options, misc = {})
      @reader  = misc[:ldap_reader] || get_reader(options)
      @rest    = misc[:rest_user_manager] || HTH::RestUserManager.new(options)
      @verbose = options[:verbose] || false
      @options = options
      @errors  = {
        update: [],
        delete: [],
        create: []
      }
    end

    def sync
      if %w(keep_in_sync sync_all).include?(@options[:strategy].to_s)
        method(@options[:strategy]).call
      else
        raise "Unknow synchronization strategy, please revise your configuration."
      end
    end

    def keep_in_sync
      hth_users = get_hth_users(@rest.find_all(only_active: true))
      source_users = @reader.find_users_by_ids(hth_users.map {|u| u[:id] })

      users_to_update = []
      users_to_delete = hth_users_no_longer_in_source(hth_users, source_users)

      source_users.each do |source_user|
        source_user = add_source(source_user)
        if source_user_active?(source_user)
          users_to_update << source_user
        else
          users_to_delete << source_user[:id]
        end
      end

      log '--- Keep Helix TeamHub users in sync ---'
      log "Users from Helix TeamHub: #{hth_users.size}"
      log "Users to update: #{users_to_update.size}"
      log "Users to deactivate: #{users_to_delete.size}"

      unless users_to_update.empty?
        perform_batch_action(users_to_update, :update)
      end

      unless users_to_delete.empty?
        perform_batch_action(users_to_delete, :delete)
      end

      group_sync
    end

    def sync_all
      source_users = @reader.find_users
      hth_users = get_hth_users(@rest.find_all(only_active: true))

      users_to_create = []
      users_to_update = []
      users_to_delete = hth_users_no_longer_in_source(hth_users, source_users)

      source_users.each do |source_user|
        source_user = add_source(source_user)
        hth_user  = get_hth_user(hth_users, source_user[:id])

        if hth_user
          if source_user_active?(source_user)
            users_to_update << source_user
          else
            users_to_delete << source_user[:id]
          end
        else
          if source_user_active?(source_user)
            users_to_create << source_user
          end
        end
      end

      log '--- Sync all LDAP users ---'
      log "Users from source: #{source_users.size}"
      log "Users from Helix TeamHub: #{hth_users.size}"
      log "Users to create: #{users_to_create.size}"
      log "Users to update: #{users_to_update.size}"
      log "Users to deactivate: #{users_to_delete.size}"

      unless users_to_create.empty?
        perform_batch_action(users_to_create, :create)
      end

      unless users_to_update.empty?
        perform_batch_action(users_to_update, :update)
      end

      unless users_to_delete.empty?
        perform_batch_action(users_to_delete, :delete)
      end

      group_sync
    end

    protected

      def group_sync
        if @options[:group_sync]
          hth_groups = get_hth_groups(@rest.find_all_groups)
          ldap_groups = @reader.get_ldap_groups

          # Making sure validations are met
          ldap_groups.each do |group|
            group[:id] = group[:id].gsub(/[^A-Za-z0-9\-_]/, '')[0..39]
            group[:name] = group[:name][0..39]
          end

          new_groups = find_new_groups(hth_groups, ldap_groups)

          log "\n--- Create groups to Helix TeamHub ---"
          log "Groups from source: #{ldap_groups.size}"
          log "Groups from Helix TeamHub: #{hth_groups.size}"
          log "Groups to create: #{new_groups.size}"

          unless new_groups.empty?
            perform_group_batch_action(new_groups, :create)
          end
        end
      end

      def get_hth_users(results)
        if results[:status] != 200
          raise "Failed to fetch users from API #{results[:content]}"
        end
        filter_users_by_source(results[:content])
      end

      def get_hth_groups(results)
        if results[:status] != 200
          raise "Failed to fetch groups from API #{results[:content]}"
        end
        results[:content]
      end

      # ensure source matches the config
      def filter_users_by_source(users)
        return users  if @options[:force_sync]
        users.find_all do |u|
          u[:source] == @options[:source]
        end
      end

      def hth_users_no_longer_in_source(hth_users, source_users)
        return []  if @options[:force_sync]
        active_hth_users = hth_users.select{ |u| u[:deleted_at].nil? }

        hth_ids  = active_hth_users.map{ |u| u[:id] }
        source_ids = source_users.map{ |u| u[:id] }

        hth_ids - source_ids
      end

      def get_hth_user(hth_users, identifier)
        hth_users.find {|user| user[:id] == identifier }
      end

      def source_user_active?(user)
        user[:active]
      end

      def add_source(source_user)
        source_user[:source] = @options[:source]
        source_user
      end

      def perform_batch_action(users, action)
        log "\n#{action} users".upcase

        if @options[:apply]
          results = @rest.send("batch_#{action}", users)
          collect_errors(results, action)
        else
          log_sanitized users
        end

        results || ""
      end

      def perform_group_batch_action(groups, action)
        log "\n#{action} groups".upcase

        if @options[:apply]
          results = @rest.send("batch_group_#{action}", groups)
          collect_errors(results, action)
        else
          log_sanitized groups
        end

        results || ""
      end

      def collect_errors(results, method)
        status_code = [:read, :update, :delete].include?(method) ? 200 : 201

        if results.nil?
          log "Empty response"
          return
        end

        results.each do |r|
          unless [200,201].include? r[:status]
            log "#{method} request failed with status #{r[:status]}:"

            unless r[:content]
              log "Response does not contain content"
              return
            end

            r[:content] = [r[:content]] unless r[:content].is_a? Array
            r[:content].each do |user|
              if user[:api_status] != status_code
                @errors[method] << sanitize_user(user)
                log "Failed to #{method}: #{user[:id]}"
              end
            end
          end
        end
      end

      def get_reader(options)
        if options[:reader] == 'ldap'
          HTH::Ldap::Reader.new(options)
        else
          abort("Invalid source given: #{options[:reader]}")
        end
      end

      def log(msg)
        puts msg if @verbose
      end

      def sanitize_user(user)
        if user.is_a? Hash
          user[:password] = "***FILTERED***"  if user[:password]
        end
        user
      end

      def log_sanitized(users)
        HTH::Utils.show(users.map { |user| sanitize_user(user) })
      end

      def find_new_groups(hth_groups, ldap_groups)
        ldap_groups.select do |ldap|
          hth_groups.select { |hth| hth[:id] == ldap[:id] }.empty?
        end
      end
  end
end

# vi: sw=2 ts=2 et ft=ruby
