require 'spec_helper'
require 'net-ldap'

describe HTH::Ldap::Reader do

  def make_ldap_entry(attributes)
    entry = Net::LDAP::Entry.new('CN=Chuck,CN=Users,DC=hth,DC=local')
    attributes.each {|key, value| entry[key] = value }
    entry
  end

  before do
    @reader_options = {
      ldap_batch_size: 2,
      ldap_required_attributes: {
        short_name: 'uid',
        email:      'mail',
        first_name: 'givenname',
        last_name:  'sn'
      },
      ldap_optional_attributes: {
        phone:       'telnumber',
        title:       'title',
        description: 'desc',
        password:    'userpasswrd',
        ldap_groups: 'memberof'
      },
      disabled_status: {
        type:      'normal',
        attribute: :employeeType,
        value:     'OBSOLETE'
      }
    }
  end

  let(:ldap)        { stub('Net::LDAP') }

  let(:connection)  { stub('Net::LDAP') }

  let(:ldap_status) { OpenStruct.new(code: 0, message: 'Success') }

  let(:ldap_reader) do
    reader = HTH::Ldap::Reader.new(@reader_options)
    reader.stubs(:ldap).returns(ldap)
    connection.stubs(:get_operation_result).returns(ldap_status)
    ldap.stubs(:open).yields(connection)
    reader
  end


  context "#find_users" do
    it "finds also groups for users" do
      connection.expects(:search).once.returns(%w(foo bar))
      ldap_reader.expects(:users_filter).once
      ldap_reader.expects(:decode_ldap_entry).with('foo').returns({short_name: :foo})
      ldap_reader.expects(:decode_ldap_entry).with('bar').returns({short_name: :bar})
      ldap_reader.expects(:ignored_user?).with(:foo).returns(false)
      ldap_reader.expects(:ignored_user?).with(:bar).returns(true)
      ldap_reader.expects(:find_groups_for_user).with('foo', connection).returns(%w(foos))
      ldap_reader.find_users.should eq([{short_name: :foo, ldap_groups: %w(foos)}])
    end

    context "when ldap operation fails" do
      let(:ldap_status) { OpenStruct.new(code: 32, message: 'No Such Object') }

      it "raises error" do
        connection.expects(:search).once.returns(nil)
        expect {ldap_reader.find_users}.to raise_error(/users.*no such object/i)
      end
    end
  end


  context "#find_users_by_ids" do
    it "searches once when batch size is not exceeded" do
      connection.expects(:search).once.returns([])
      ldap_reader.expects(:short_names_filter).with(%w(a b)).once
      ldap_reader.find_users_by_ids(%w(a b))
    end

    it "searches in batches when needed" do
      connection.expects(:search).times(3).returns([])
      ldap_reader.expects(:short_names_filter).with(%w(a b)).once
      ldap_reader.expects(:short_names_filter).with(%w(c d)).once
      ldap_reader.expects(:short_names_filter).with(%w(e)).once
      ldap_reader.find_users_by_ids(%w(a b c d e))
    end

    it "finds groups for users" do
      connection.expects(:search).once.returns(%w(foo bar))
      ldap_reader.expects(:decode_ldap_entry).with('foo').returns({short_name: :foo})
      ldap_reader.expects(:decode_ldap_entry).with('bar').returns({short_name: :bar})
      ldap_reader.expects(:find_groups_for_user).with('foo', connection).returns(%w(foos))
      ldap_reader.expects(:find_groups_for_user).with('bar', connection).returns(%w(bars))
      ldap_reader.find_users_by_ids(%w(foo bar)).should eq([
        {short_name: :foo, ldap_groups: %w(foos)},
        {short_name: :bar, ldap_groups: %w(bars)}
      ])
    end

    context "when ldap operation fails" do
      let(:ldap_status) { OpenStruct.new(code: 32, message: 'No Such Object') }

      it "raises error" do
        connection.expects(:search).once.returns(nil)
        ldap_reader.expects(:short_names_filter).with(%w(a b)).once
        expect {ldap_reader.find_users_by_ids(%w(a b))}.to raise_error(/users.*no such object/i)
      end
    end
  end


  context "#find_groups_for_user" do
    let(:ldap_entry_attributes) do
      {
        'memberof' => ["HTH_USERS", "cn=I_DEVS,ou=groups"]
      }
    end

    let(:user_entry) { make_ldap_entry(ldap_entry_attributes) }

    let(:connection) { stub('Net::LDAP') }

    let(:user_ldap_groups) { ldap_reader.find_groups_for_user(user_entry, connection) }

    context "when available from user entries" do
      before do
        connection.expects(:search).never
      end

      context "when user is member of groups" do
        it "returns DN and CN of groups" do
          user_ldap_groups.should eq ["HTH_USERS", "cn=I_DEVS,ou=groups", "I_DEVS"]
        end
      end

      context "when user is not a member of any group (attribute not present)" do
        let(:ldap_entry_attributes) { {} }

        it "returns empty array" do
          user_ldap_groups.should eq []
        end
      end

      context "when empty" do
        let(:ldap_entry_attributes) { {'memberof' => []} }

        it "returns empty array" do
          user_ldap_groups.should eq []
        end
      end

      context "when not mapped" do
        before do
          @reader_options[:ldap_optional_attributes].delete(:ldap_groups)
        end

        it "returns empty array" do
          user_ldap_groups.should eq []
        end
      end
    end

    context "when requested separately" do
      before do
        @reader_options[:request_user_groups] = true
        @reader_options[:group_base] = 'ou=groups,o=foobar'
        @reader_options[:ldap_optional_attributes][:ldap_groups] = 'member'
        connection.stubs(:search).with(expected_search_params).returns(group_entries)
      end

      let(:expected_search_params) do
        {
          return_result: true,
          base: 'ou=groups,o=foobar',
          filter: Net::LDAP::Filter.eq('member', 'CN=Chuck,CN=Users,DC=hth,DC=local'),
          attributes: %w(dn)
        }
      end

      let(:group_entries) do
        [
          Net::LDAP::Entry.new('CN=developers,o=foobar'),
          Net::LDAP::Entry.new('CN=designers,o=foobar')
        ]
      end

      context "when user is a member of groups" do
        it "returns DN and CN of groups" do
          user_ldap_groups.should eq(group_entries.map(&:dn) + %w(developers designers))
        end
      end

      context "when user is not a member of any group" do
        let(:group_entries) { [] }

        it "returns empty list" do
          user_ldap_groups.should eq []
        end
      end

      context "when not mapped correctly" do
        before do
          @reader_options[:ldap_optional_attributes][:ldap_groups] = 'memberof'
        end

        it "raises error" do
          expect{ user_ldap_groups }.to raise_error(/memberof/i)
        end
      end

      context "when ldap operation fails" do
        let(:group_entries) { nil }
        let(:ldap_status) { OpenStruct.new(code: 32, message: 'No Such Object') }

        it "raises error" do
          expect {
            user_ldap_groups
          }.to raise_error(/groups.*no such object/i)
        end
      end
    end
  end


  context "#ldap_attribute_names" do
    it "contains required attributes" do
      ldap_reader.ldap_attribute_names.should include('uid')
      ldap_reader.ldap_attribute_names.should include('mail')
    end

    it "contains optional attributes" do
      ldap_reader.ldap_attribute_names.should include('memberof')
      ldap_reader.ldap_attribute_names.should include('userpasswrd')
    end

    it "contains status attribute" do
      ldap_reader.ldap_attribute_names.should include('employeeType')
    end

    context "when optional attributes are missing" do
      before do
        @reader_options.delete(:disabled_status)
        @reader_options.delete(:ldap_optional_attributes)
      end

      it "doesn't break" do
        ldap_reader.ldap_attribute_names.should_not include('employeeType')
        ldap_reader.ldap_attribute_names.should_not include('memberof')
      end
    end
  end


  context "#users_filter" do
    context "without user defined filter" do
      it "filters only by required attributes" do
        expected = '(&(&(&(uid=*)(givenname=*))(sn=*))(mail=*))'
        ldap_reader.users_filter.to_s.should eq expected
      end
    end

    context "with user defined filter" do
      before do
        @reader_options = @reader_options.merge(
          query_filters: '(objectClass=user)'
        )
      end

      it "filters by filter AND required attributes" do
        expected = '(&(objectClass=user)(&(&(&(uid=*)(givenname=*))(sn=*))(mail=*)))'
        ldap_reader.users_filter.to_s.should eq expected
      end
    end
  end


  context "#short_names_filter" do
    context "without user defined filter" do
      it "filters only by list of possible short_names" do
        expected = '(|(|(uid=a)(uid=b))(uid=c))'
        ldap_reader.short_names_filter(%w(a b c)).to_s.should eq expected
      end
    end

    context "with user defined filter" do
      before do
        @reader_options = @reader_options.merge(
          query_filters: '(objectClass=user)'
        )
      end

      it "filters by filter AND list of possible short_names" do
        expected = '(&(objectClass=user)(|(|(uid=a)(uid=b))(uid=c)))'
        ldap_reader.short_names_filter(%w(a b c)).to_s.should eq expected
      end
    end
  end


  context "#group_member_filter" do
    [
      {
        attribute: 'member',
        filter: "(member=CN=Chuck,CN=Users,DC=hth,DC=local)"
      },
      {
        attribute: 'member:1.2.840.113556.1.4.1941:',
        filter: "(member:1.2.840.113556.1.4.1941:=CN=Chuck,CN=Users,DC=hth,DC=local)"
      },
      {
        attribute: 'uniqueMember',
        filter: "(uniqueMember=CN=Chuck,CN=Users,DC=hth,DC=local)"
      },
      {
        attribute: 'memberUid',
        filter: "(memberUid=chuck)"
      }
    ].each do |hash|
      [:to_s, :upcase].each do |method| # test case-insensitive
        context "when #{hash[:attribute].send(method)} attribute" do
          before do
            @reader_options[:ldap_optional_attributes][:ldap_groups] = hash[:attribute].send(method)
          end

          it "returns correct filter #{hash[:filter]}" do
            ldap_reader.group_member_filter(make_ldap_entry({uid: 'chuck'})).to_s.should eq(hash[:filter])
          end
        end
      end
    end
  end


  context "reading ldap values" do
    let(:entry) {
      make_ldap_entry({
        'uid'  => ["ckent", "asdf"],
        'mail' => []
      })
    }

    context "#get_single_value" do
      it "returns first value" do
        ldap_reader.get_single_value(entry, :uid).should eq 'ckent'
      end

      it "returns nil when there are no values" do
        ldap_reader.get_single_value(entry, :mail).should eq nil
      end

      it "returns nil for nonexisting attribute" do
        ldap_reader.get_single_value(entry, :title).should eq nil
      end
    end

    context "#get_multi_value" do
      it "returns all values" do
        ldap_reader.get_multi_value(entry, :uid).should eq ["ckent", "asdf"]
      end

      it "returns empty array when there are no values" do
        ldap_reader.get_multi_value(entry, :mail).should eq []
      end

      it "returns empty array for nonexisting attribute" do
        ldap_reader.get_multi_value(entry, :title).should eq []
      end
    end

    context "encoding" do
      let(:values) {
        ["f\xC3\xB6\xC3\xB6", "b\xC3\xA4r"].map {|v| v.force_encoding('ASCII-8BIT') }
      }
      let(:entry) {
        make_ldap_entry({
          'uid' => values
        })
      }

      it "is forced to UTF-8 for single values" do
        ldap_reader.get_single_value(entry, :uid).encoding.name.should eq "UTF-8"
      end

      it "is forced to UTF-8 for multi values" do
        encodings = ldap_reader.get_multi_value(entry, :uid).map(&:encoding)
        encodings.map(&:name).uniq.should eq ["UTF-8"]
      end
    end
  end


  context "detecting pre hashed passwords" do
    before do
      @ldap_sha1_password = "{SHA}C+7Hteo/D9vJXQ3UfzxbwnXaijM="
      @ldap_md5_password  = "{MD5}rL0Y20zC+Fzt72VPzMSk2A=="
      @plaintext_password = "foo"
    end

    it "should match sha1" do
      ldap_reader.password_pre_hash(@ldap_sha1_password).should eq :ldap_sha1
    end

    it "should match md5" do
      ldap_reader.password_pre_hash(@ldap_md5_password).should eq :ldap_md5
    end

    it "should match plaintext" do
      ldap_reader.password_pre_hash(@plaintext_password).should eq nil
    end
  end


  context "detecting active users" do
    context "when attributes missing" do
      before do
        @reader_options.delete(:disabled_status)
      end

      it "returns true" do
        ldap_reader.active_status(make_ldap_entry({})).should be_true
      end
    end

    context "based on bitmask attribute" do
      before do
        @reader_options = @reader_options.merge({
          disabled_status: {
            type:      'bitmask',
            attribute: :useraccountcontrol,
            value:      2
          }
        })
      end

      let(:active_entry)   { make_ldap_entry({useraccountcontrol: [512]}) }
      let(:inactive_entry) { make_ldap_entry({useraccountcontrol: [514]}) }

      it "should return false for inactive entry" do
        ldap_reader.active_status(inactive_entry).should be_false
      end

      it "should return true for active entry" do
        ldap_reader.active_status(active_entry).should be_true
      end

      it "should return true when status is unknown" do
        ldap_reader.active_status(make_ldap_entry({})).should be_true
      end
    end

    context "based on normal attribute" do
      before do
        @reader_options = @reader_options.merge({
          disabled_status: {
            type:      'normal',
            attribute: :employeeType,
            value:      'OBSOLETE'
          }
        })
      end

      let(:active_entry)   { make_ldap_entry({employeeType: ['ACTIVE', 'INTERNAL']}) }
      let(:inactive_entry) { make_ldap_entry({employeeType: ['INTERNAL', 'OBSOLETE']}) }

      it "should return false for inactive entry" do
        ldap_reader.active_status(inactive_entry).should be_false
      end

      it "should return true for active entry" do
        ldap_reader.active_status(active_entry).should be_true
      end

      it "should return true when status is unknown" do
        ldap_reader.active_status(make_ldap_entry({})).should be_true
      end
    end
  end


  context "decoding ldap entries" do
    let(:ldap_entry_attributes) do
      {
        'uid'            => ["ckent", "asdf"],
        'mail'           => ["ckent@foo.com"],
        'telnumber'      => ["12345678"],
        'title'          => [],
      }
    end

    let(:decoded) {
      ldap_reader.decode_ldap_entry(make_ldap_entry(ldap_entry_attributes))
    }

    it "has required attributes" do
      decoded[:short_name].should eq 'ckent'
      decoded[:email].should      eq 'ckent@foo.com'
    end

    it "has optional attributes with values" do
      decoded[:phone].should eq '12345678'
    end

    it "does not have optional attributes without values" do
      decoded.should_not have_key :title
    end

    it "does not have nonexisting attributes" do
      decoded.should_not have_key :description
    end

    it "keeps track of synchronized fields" do
      decoded[:synchronized_fields].should =~ %w{short_name email phone}
    end

    it "adds short_name as id" do
      decoded[:id].should eq 'ckent'
    end

    it "is active by default" do
      decoded[:active].should be_true
    end

    context "when password is present" do
      let(:ldap_entry_attributes) do
        {
          'userpasswrd' => ["{MD5}rL0Y20zC+Fzt72VPzMSk2A=="]
        }
      end

      it "is also synchronized with related fields" do
        decoded[:synchronized_fields].should include("password")
        decoded[:synchronized_fields].should include("pre_hash")
      end

      it "has correct pre_hash" do
        decoded[:pre_hash].should eq :ldap_md5
      end
    end
  end
end
